@@ -1,10 +1,12 @@
-import numpy as np
-import photutils
 import astropy.units as u
+import numpy as np
+from photutils.aperture import EllipticalAperture, RectangularAperture
+from scipy.ndimage import fourier_shift
+from skimage.transform import rescale, resize
 
 from paos import logger
-from paos.classes.zernike import Zernike
 from paos.classes.psd import PSD
+from paos.classes.zernike import PolyOrthoNorm, Zernike
 
 
 class WFO:
@@ -66,6 +68,8 @@ class WFO:
     extent: tuple
         the physical coordinates of the wavefront bounding box (xmin, xmax, ymin, ymax).
         Can be used directly in im.set_extent.
+    propagator: string
+        Propagator type last applied
 
     Returns
     -------
@@ -93,7 +97,6 @@ class WFO:
     """
 
     def __init__(self, beam_diameter, wl, grid_size, zoom):
-
         assert np.log2(grid_size).is_integer(), "Grid size should be 2**n"
         assert zoom > 0, "zoom factor should be positive"
         assert beam_diameter > 0, "beam diameter should be positive"
@@ -113,6 +116,8 @@ class WFO:
 
         grid_size = np.uint(grid_size)
         self._wfo = np.ones((grid_size, grid_size), dtype=np.complex128)
+        self._zoom = zoom
+        self._propagator = ""
 
     @property
     def wl(self):
@@ -174,6 +179,10 @@ class WFO:
     def distancetofocus(self):
         return self.zw0 - self.z
 
+    @property
+    def propagator(self):
+        return self._propagator
+
     @property
     def extent(self):
         return (
@@ -234,9 +243,7 @@ class WFO:
             ihx = hx / self.dx
             ihy = hy / self.dy
             theta = 0.0 if tilt is None else np.deg2rad(tilt)
-            aperture = photutils.aperture.EllipticalAperture(
-                (ixc, iyc), ihx, ihy, theta=theta
-            )
+            aperture = EllipticalAperture((ixc, iyc), ihx, ihy, theta=theta)
             mask = aperture.to_mask(method="exact").to_image(self._wfo.shape)
         elif shape == "circular":
             if r is None:
@@ -245,9 +252,7 @@ class WFO:
             ihx = r / self.dx
             ihy = r / self.dy
             theta = 0.0
-            aperture = photutils.aperture.EllipticalAperture(
-                (ixc, iyc), ihx, ihy, theta=theta
-            )
+            aperture = EllipticalAperture((ixc, iyc), ihx, ihy, theta=theta)
             mask = aperture.to_mask(method="exact").to_image(self._wfo.shape)
         elif shape == "rectangular":
             if hx is None or hy is None:
@@ -256,16 +261,14 @@ class WFO:
             ihx = hx / self.dx
             ihy = hy / self.dy
             theta = 0.0 if tilt is None else np.deg2rad(tilt)
-            aperture = photutils.aperture.RectangularAperture(
-                (ixc, iyc), ihx, ihy, theta=theta
-            )
+            aperture = RectangularAperture((ixc, iyc), ihx, ihy, theta=theta)
             # Exact method not implemented in photutils 1.0.2
             mask = aperture.to_mask(method="subpixel", subpixels=32).to_image(
                 self._wfo.shape
             )
         else:
-            logger.error("Aperture {:s} not defined yet.".format(shape))
-            raise ValueError("Aperture {:s} not defined yet.".format(shape))
+            logger.error(f"Aperture {shape:s} not defined yet.")
+            raise ValueError(f"Aperture {shape:s} not defined yet.")
 
         if obscuration:
             self._wfo *= 1 - mask
@@ -325,9 +328,7 @@ class WFO:
 
         # update Gaussian beam parameters
         self._w0 = wz / np.sqrt(1.0 + (np.pi * wz**2 * gCima / self.wl) ** 2)
-        self._zw0 = (
-            -gCima / (gCima**2 + (self.wl / (np.pi * wz**2)) ** 2) + self.z
-        )
+        self._zw0 = -gCima / (gCima**2 + (self.wl / (np.pi * wz**2)) ** 2) + self.z
         self._zr = np.pi * self.w0**2 / self.wl
 
         propagator = propagator + self.insideout()
@@ -392,9 +393,7 @@ class WFO:
         self._dy *= My
 
         if np.abs(Mx - 1.0) < 1.0e-8 or Mx is None:
-            logger.trace(
-                "Does not do anything if magnification x is close to unity."
-            )
+            logger.trace("Does not do anything if magnification x is close to unity.")
             return
 
         logger.warning(
@@ -453,9 +452,7 @@ class WFO:
             propagation distance
         """
         if np.abs(dz) < 0.001 * self.wl:
-            logger.debug(
-                "Thickness smaller than 1/1000 wavelength. Returning.."
-            )
+            logger.debug("Thickness smaller than 1/1000 wavelength. Returning..")
             return
 
         if self.C != 0:
@@ -484,9 +481,7 @@ class WFO:
             propagation distance
         """
         if np.abs(dz) < 0.001 * self.wl:
-            logger.debug(
-                "Thickness smaller than 1/1000 wavelength. Returning.."
-            )
+            logger.debug("Thickness smaller than 1/1000 wavelength. Returning..")
             return
 
         if self.C == 0.0:
@@ -523,9 +518,7 @@ class WFO:
             propagation distance
         """
         if np.abs(dz) < 0.001 * self.wl:
-            logger.debug(
-                "Thickness smaller than 1/1000 wavelength. Returning.."
-            )
+            logger.debug("Thickness smaller than 1/1000 wavelength. Returning..")
             return
 
         if self.C != 0.0:
@@ -548,7 +541,7 @@ class WFO:
         self._z = self._z + dz
         self._C = 1 / (self.z - self.zw0)
         self._dx = self.wl * np.abs(dz) / (wf.shape[1] * self.dx)
-        self._dy = self.wl * np.abs(dz) / (wf.shape[1] * self.dy)
+        self._dy = self.wl * np.abs(dz) / (wf.shape[0] * self.dy)
         self._wfo = np.fft.fftshift(wf)
 
     def propagate(self, dz):
@@ -576,11 +569,23 @@ class WFO:
             self.stw(self.zw0 - z1)
             self.wts(z2 - self.zw0)
 
+        self._propagator = propagator
+
     def zernikes(
-        self, index, Z, ordering, normalize, radius, offset=0.0, origin="x"
+        self,
+        index,
+        Z,
+        ordering,
+        normalize,
+        radius,
+        offset=0.0,
+        origin="x",
+        orthonorm=False,
+        mask=False,
     ):
         """
-        Add a WFE represented by a Zernike expansion
+        Add a WFE represented by a Zernike expansion.
+        Either Zernike or orthonormal polynomials can be used.
 
         Parameters
         ----------
@@ -599,21 +604,28 @@ class WFO:
         origin: string
             Angles measured counter-clockwise positive from x axis by default (origin='x').
             Set origin='y' for angles measured clockwise-positive from the y-axis.
+        orthonorm: bool
+            If True, orthonormal polynomials are used. Default is False.
+        mask: bool array like
+            The mask defining the pupil following masked array convention.
+            Pixel within the pupil are masked False. Defaults to False.
 
         Returns
         -------
         out: masked array
             the WFE
         """
-        assert not np.any(
-            np.diff(index) - 1
-        ), "Zernike sequence should be continuous"
+        assert not np.any(np.diff(index) - 1), "Zernike sequence should be continuous"
 
         x = (np.arange(self._wfo.shape[1]) - self._wfo.shape[1] // 2) * self.dx
         y = (np.arange(self._wfo.shape[0]) - self._wfo.shape[0] // 2) * self.dy
 
         xx, yy = np.meshgrid(x, y)
-        rho = np.sqrt(xx**2 + yy**2) / radius
+        rho = np.ma.MaskedArray(
+            data=np.sqrt(xx**2 + yy**2) / radius,
+            mask=mask,
+            fill_value=0.0,
+        )
 
         if origin == "x":
             phi = np.arctan2(yy, xx) + np.deg2rad(offset)
@@ -621,26 +633,243 @@ class WFO:
             phi = np.arctan2(xx, yy) + np.deg2rad(offset)
         else:
             logger.error(
-                "Origin {} not recognised. Origin shall be either x or y".format(
-                    origin
-                )
+                f"Origin {origin} not recognised. Origin shall be either x or y"
             )
             raise ValueError(
-                "Origin {} not recognised. Origin shall be either x or y".format(
-                    origin
-                )
+                f"Origin {origin} not recognised. Origin shall be either x or y"
             )
-        zernike = Zernike(
-            len(index), rho, phi, ordering=ordering, normalize=normalize
-        )
+
+        func = PolyOrthoNorm if orthonorm else Zernike
+        logger.debug(f"Using {func.__name__} polynomials")
+
+        zernike = func(len(index), rho, phi, ordering=ordering, normalize=normalize)
         zer = zernike()
         wfe = (zer.T * Z).T.sum(axis=0)
+        logger.debug(f"WFE RMS = {np.std(wfe)}")
+
         self._wfo = self._wfo * np.exp(
-            2.0 * np.pi * 1j * wfe / self._wl
-        ).filled(0)
+            2.0 * np.pi * 1j * wfe.filled(0) / self._wl
+        )  # Note: applied filled(0) inside the exponent rather than outside to avoid cropping wavefront outside of the wfe mask
 
         return wfe
 
+    def grid_sag(
+        self,
+        sag: np.ndarray,
+        nx: int,
+        ny: int,
+        delx: float,
+        dely: float,
+        xdec: float = 0.0,
+        ydec: float = 0.0,
+    ):
+        """
+        Add a user-specified grid sag to the wavefront
+
+        Parameters
+        ----------
+        sag: array
+            2D array of sag values in meters
+        nx: int
+            Number of sag pixels along x-axis.
+        ny: int
+            Number of sag pixels along y-axis.
+        delx: float
+            pixel size along x-axis. The user can specify a pixel size
+            in the sag array to account for a possible difference
+            between the sag and the WFO.
+        dely: float
+            pixel size along y-axis. See delx.
+        xdec: float
+            pixel shift along x-axis. The user can specify a pixel shift
+            in the sag array to account for a possible misalignment
+            between the sag and the WFO.
+        ydec: float
+            pixel shift along y-axis. See xdec.
+
+        Returns
+        -------
+        out: array
+            the WFE
+        """
+
+        def rescale_map(sag, mask, scale_x, scale_y):
+
+            anti_aliasing = (
+                scale_x < 1.0 or scale_y < 1.0
+            )  # anti_aliasing is required for downsampling
+
+            sag = rescale(
+                sag,
+                scale=(scale_y, scale_x),
+                anti_aliasing=anti_aliasing,
+                order=3,
+            )
+
+            mask = rescale(
+                mask,
+                scale=(scale_y, scale_x),
+                anti_aliasing=anti_aliasing,
+                order=3,
+            )
+
+            return sag, mask
+
+        def pad_map(sag, mask, padding):
+            sag = np.pad(
+                sag,
+                padding,
+                mode="constant",
+                constant_values=0,
+            )
+            mask = np.pad(
+                mask,
+                padding,
+                mode="constant",
+                constant_values=1,
+            )
+            return sag, mask
+
+        def resize_map(sag, mask, scale_x, scale_y):
+
+            anti_aliasing = (
+                scale_x < 1.0 or scale_y < 1.0
+            )  # anti_aliasing is required for downsampling
+
+            sag = resize(
+                sag,
+                output_shape=self._wfo.shape,
+                anti_aliasing=anti_aliasing,
+                order=3,
+            )
+            mask = resize(
+                mask,
+                output_shape=self._wfo.shape,
+                anti_aliasing=anti_aliasing,
+                order=3,
+            )
+            return sag, mask
+
+        assert sag.ndim == 2, "sag shall be a 2D array"
+        assert sag.shape == (ny, nx)
+
+        logger.debug("Converting sag to masked array")
+        if isinstance(sag, np.ma.MaskedArray):
+            logger.debug("Input sag is already a masked array")
+        else:
+            logger.debug("Input sag is not a masked array")
+            mask = ~np.isfinite(sag) | (sag == 0)
+            sag = np.ma.MaskedArray(sag, mask=mask)
+
+        mask = sag.mask.astype(float)
+        sag = sag.filled(0.0)
+
+        # Step 1: recenter
+        if (xdec != 0) or (ydec != 0):
+            logger.debug("Applying sag shift: xdec = %f, ydec = %f" % (xdec, ydec))
+            sag = fourier_shift(np.fft.fft2(sag), shift=(-xdec, -ydec))
+            sag = np.fft.ifft2(sag).real
+            mask = fourier_shift(np.fft.fft2(mask), shift=(-xdec, -ydec))
+            mask = np.fft.ifft2(mask).real
+
+        # Step 2: pad or crop
+        target_width = self._wfo.shape[1] * self._dx
+        target_height = self._wfo.shape[0] * self._dy
+
+        current_width = sag.shape[1] * delx
+        current_height = sag.shape[0] * dely
+
+        logger.debug(
+            f"target width [m]: {target_width}, target height [m]: {target_height}"
+        )
+        logger.debug(
+            f"current width [m]: {current_width}, current height [m]: {current_height}"
+        )
+
+        width_diff = int(np.floor((current_width - target_width) / delx))
+        height_diff = int(np.floor((current_height - target_height) / dely))
+
+        scale_x = scale_y = 1
+        if width_diff % 2 == 1 or height_diff % 2 == 1:
+            logger.debug(f"I need to sample more finely. Sag shape is {sag.shape}")
+
+        if width_diff % 2 == 1:
+            scale_x = 2
+            delx /= 2
+            width_diff *= 2
+
+        if height_diff % 2 == 1:
+            scale_y = 2
+            dely /= 2
+            height_diff *= 2
+
+        if (scale_x != 1) or (scale_y != 1):
+            sag, mask = rescale_map(sag, mask, scale_x, scale_y)
+        logger.debug(f"Resampled sag shape is {sag.shape}")
+
+        # Handle width dimension (x-axis)
+        if width_diff < 0.0:
+            logger.debug("Applying padding on width...")
+            pad_width = abs(width_diff)
+            pad_left = pad_width // 2
+            pad_right = pad_width - pad_left
+            sag, mask = pad_map(sag, mask, ((0, 0), (pad_left, pad_right)))
+
+        elif width_diff > 0.0:
+            logger.debug("Applying cropping on width...")
+            crop_width = width_diff
+            crop_left = crop_width // 2
+            crop_right = sag.shape[1] - (crop_width - crop_left)
+
+            # Apply cropping only on x-axis
+            sag = sag[:, crop_left:crop_right]
+            mask = mask[:, crop_left:crop_right]
+
+        # Handle height dimension (y-axis)
+        if height_diff < 0.0:
+            logger.debug("Applying padding on height...")
+            pad_height = abs(height_diff)
+            pad_top = pad_height // 2
+            pad_bottom = pad_height - pad_top
+            sag, mask = pad_map(sag, mask, ((pad_top, pad_bottom), (0, 0)))
+
+        elif height_diff > 0.0:
+            logger.debug("Applying cropping on height...")
+            crop_height = height_diff
+            crop_top = crop_height // 2
+            crop_bottom = sag.shape[0] - (crop_height - crop_top)
+
+            # Apply cropping only on y-axis
+            sag = sag[crop_top:crop_bottom, :]
+            mask = mask[crop_top:crop_bottom, :]
+        logger.debug(f"Adjusted sag shape is {sag.shape}, mask shape is {mask.shape}")
+
+        # Step 3: rescale
+        scale_x = delx / self.dx
+        scale_y = dely / self.dy
+
+        if (scale_x != 1) or (scale_y != 1):
+            sag, mask = rescale_map(sag, mask, scale_x, scale_y)
+
+        # Step 4: resize
+        # if the shape is not the same as the input (could be 1 pixel off), resize
+        logger.debug(f"Resampled sag shape is {sag.shape}")
+
+        if sag.shape != self._wfo.shape:
+            logger.debug(f"Output shape should be {self._wfo.shape}: resizing...")
+            scale_x = self._wfo.shape[1] / sag.shape[1]
+            scale_y = self._wfo.shape[0] / sag.shape[0]
+            sag, mask = resize_map(sag, mask, scale_x, scale_y)
+
+        mask = mask > 0.1
+        sag = np.ma.MaskedArray(sag, mask=mask)
+
+        self._wfo = self._wfo * np.exp(
+            2.0 * np.pi * 1j * sag.filled(0) / self._wl
+        )  # Note: applied filled(0) inside the exponent rather than outside to avoid cropping wavefront outside of the sag mask
+
+        return sag
+
     def psd(
         self,
         A=10.0,
@@ -681,22 +910,22 @@ class WFO:
         """
 
         # compute 2D frequency grid
-        fx = np.fft.fftfreq(self._wfo.shape[0], self.dx)
-        fy = np.fft.fftfreq(self._wfo.shape[1], self.dy)
+        fx = np.fft.fftfreq(self._wfo.shape[1], self.dx)
+        fy = np.fft.fftfreq(self._wfo.shape[0], self.dy)
 
         fxx, fyy = np.meshgrid(fx, fy)
         f = np.sqrt(fxx**2 + fyy**2)
         f[f == 0] = 1e-100
 
         if fmax is None:
-            print("WARNING: fmax not provided, using f_Nyq")
+            logger.warning("fmax not provided, using f_Nyq")
             fmax = 0.5 * np.sqrt(self.dx**-2 + self.dy**-2)
         else:
             f_Nyq = 0.5 * np.sqrt(self.dx**-2 + self.dy**-2)
             assert fmax <= f_Nyq, f"fmax must be less than or equal to f_Nyq ({f_Nyq})"
 
         if fmin is None:
-            print("WARNING: fmin not provided, using 1 / D")
+            logger.warning("fmin not provided, using 1 / D")
             fmin = 1 / (self._wfo.shape[0] * np.max([self.dx, self.dy]))
 
         # compute 2D PSD
